from typing import Any, Optional, Union

from edc_constants.constants import BLACK, FEMALE, MALE, OTHER

from ..convert_units import convert_units
from ..units import MILLIGRAMS_PER_DECILITER
from .exceptions import CalculatorError

# TODO: https://www.rcpa.edu.au/Manuals/RCPA-Manual/
#  Pathology-Tests/C/Creatinine-clearance-Cockcroft-and-Gault


class eGFR:  # noqa

    """Reference https://nephron.com/epi_equation

    Levey AS, Stevens LA, et al. A New Equation to Estimate Glomerular
    Filtration Rate. Ann Intern Med. 2009; 150:604-612.
    """

    def __init__(
        self: Any,
        gender: Optional[str] = None,
        age: Optional[int] = None,
        ethnicity: Optional[str] = None,
        creatinine_value: Optional[Union[float, int]] = None,
        creatinine_units: Optional[str] = None,
    ):
        """Expects creatinine (scr) in umols/L.

        Converts to creatinine to mg/dL for the calculation.
        """

        if not gender or gender not in [MALE, FEMALE]:
            raise CalculatorError(f"Invalid gender. Expected on of {MALE}, {FEMALE}")
        self.gender = gender

        if not (18 <= (age or 0) < 120):
            raise CalculatorError(f"Invalid age. See {self.__class__.__name__}. Got {age}")
        self.age = float(age)

        self.ethnicity = ethnicity or OTHER
        self.scr = convert_units(
            float(creatinine_value),
            units_from=creatinine_units,
            units_to=MILLIGRAMS_PER_DECILITER,
        )

    @property
    def value(self):
        return (
            141.000
            * (min(self.scr / self.kappa, 1.000) ** self.alpha)
            * (max(self.scr / self.kappa, 1.000) ** -1.209)
            * self.age_factor
            * self.gender_factor
            * self.ethnicity_factor
        )

    @property
    def alpha(self):
        return -0.329 if self.gender == FEMALE else -0.411

    @property
    def kappa(self):
        return 0.7 if self.gender == FEMALE else 0.9

    @property
    def ethnicity_factor(self):
        return 1.150 if self.ethnicity == BLACK else 1.000

    @property
    def gender_factor(self):
        return 1.018 if self.gender == FEMALE else 1.000

    @property
    def age_factor(self):
        return 0.993**self.age


def calculate_egfr(
    gender: Optional[str] = None,
    age_in_years: Optional[int] = None,
    ethnicity: Optional[str] = None,
    creatinine_value: Optional[Union[float, int]] = None,
    creatinine_units: Optional[str] = None,
    **kwargs,  # noqa
):
    calculated_egfr_value = None
    if gender and age_in_years and ethnicity and creatinine_value and creatinine_units:
        opts = dict(
            gender=gender,
            age=age_in_years,
            ethnicity=ethnicity,
            creatinine_value=creatinine_value,
            creatinine_units=creatinine_units,
        )
        calculated_egfr_value = eGFR(**opts).value
    return calculated_egfr_value



from dateutil.relativedelta import relativedelta
from django.core.exceptions import ValidationError
from django.test import TestCase
from edc_constants.constants import FEMALE, NO, NOT_APPLICABLE, YES
from edc_utils import get_utcnow

from edc_reportable import (
    GRAMS_PER_DECILITER,
    IU_LITER,
    MICROMOLES_PER_LITER,
    MILLIMOLES_PER_LITER,
    TEN_X_9_PER_LITER,
    site_reportables,
)
from edc_reportable.units import MILLIGRAMS_PER_DECILITER
from reportable_app.form_validators import SpecimenResultFormValidator
from reportable_app.models import SpecimenResult
from reportable_app.reportables import grading_data, normal_data


class TestSpecimenResultForm(TestCase):
    def setUp(self):
        site_reportables._registry = {}

        site_reportables.register(
            name="my_reference_list", normal_data=normal_data, grading_data=grading_data
        )

        self.cleaned_data = {
            "subject_visit": "",
            "dob": get_utcnow() - relativedelta(years=25),
            "gender": FEMALE,
            "haemoglobin": 15,
            "haemoglobin_units": GRAMS_PER_DECILITER,
            "haemoglobin_abnormal": NO,
            "haemoglobin_reportable": NOT_APPLICABLE,
            "alt": 10,
            "alt_units": IU_LITER,
            "alt_abnormal": NO,
            "alt_reportable": NOT_APPLICABLE,
            "magnesium": 0.8,
            "magnesium_units": MILLIMOLES_PER_LITER,
            "magnesium_abnormal": NO,
            "magnesium_reportable": NOT_APPLICABLE,
            "creatinine": 100,
            "creatinine_units": MICROMOLES_PER_LITER,
            "creatinine_abnormal": NO,
            "creatinine_reportable": NOT_APPLICABLE,
            "neutrophil": 3,
            "neutrophil_units": TEN_X_9_PER_LITER,
            "neutrophil_abnormal": NO,
            "neutrophil_reportable": NOT_APPLICABLE,
            "sodium": 135,
            "sodium_units": MILLIMOLES_PER_LITER,
            "sodium_abnormal": NO,
            "sodium_reportable": NOT_APPLICABLE,
            "potassium": 4.0,
            "potassium_units": MILLIMOLES_PER_LITER,
            "potassium_abnormal": NO,
            "potassium_reportable": NOT_APPLICABLE,
            "platelets": 450,
            "platelets_units": TEN_X_9_PER_LITER,
            "platelets_abnormal": NO,
            "platelets_reportable": NOT_APPLICABLE,
            "results_normal": YES,
            "results_reportable": NOT_APPLICABLE,
        }

    def test_haemoglobin_units_invalid(self):
        self.cleaned_data.update(haemoglobin=6.4, results_abnormal=YES)
        form_validator = SpecimenResultFormValidator(
            cleaned_data=self.cleaned_data, instance=SpecimenResult()
        )
        self.assertRaises(ValidationError, form_validator.validate)
        self.assertIn("haemoglobin", form_validator._errors)

    def test_haemoglobin_units_male_valid(self):
        self.cleaned_data.update(haemoglobin=14, results_abnormal=NO)
        form_validator = SpecimenResultFormValidator(
            cleaned_data=self.cleaned_data, instance=SpecimenResult()
        )
        try:
            form_validator.validate()
        except ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got{e}")

    def test_no_creatinine_mg_invalid(self):
        self.cleaned_data.update(creatinine=0.3, creatinine_units=MILLIGRAMS_PER_DECILITER)
        form_validator = SpecimenResultFormValidator(
            cleaned_data=self.cleaned_data, instance=SpecimenResult()
        )
        self.assertRaises(ValidationError, form_validator.validate)
        self.assertIn("creatinine", form_validator._errors)

    def test_no_creatinine_mg_sodium_invalid(self):
        self.cleaned_data.update(
            creatinine=900, creatinine_units=MICROMOLES_PER_LITER, results_abnormal=YES
        )
        form_validator = SpecimenResultFormValidator(
            cleaned_data=self.cleaned_data, instance=SpecimenResult()
        )
        self.assertRaises(ValidationError, form_validator.validate)
        self.assertIn("creatinine", form_validator._errors)

    def test_creatinine_mg_invalid(self):
        self.cleaned_data.update(
            creatinine=2.48,
            creatinine_units=MILLIGRAMS_PER_DECILITER,
            results_abnormal=YES,
        )
        form_validator = SpecimenResultFormValidator(
            cleaned_data=self.cleaned_data, instance=SpecimenResult()
        )
        self.assertRaises(ValidationError, form_validator.validate)
        self.assertIn("creatinine", form_validator._errors)

    def test_creatinine_mg(self):

        self.cleaned_data.update(
            creatinine=1.3,
            creatinine_units=MILLIGRAMS_PER_DECILITER,
            results_abnormal=NO,
            results_reportable=NOT_APPLICABLE,
        )
        form_validator = SpecimenResultFormValidator(
            cleaned_data=self.cleaned_data, instance=SpecimenResult()
        )
        try:
            form_validator.validate()
        except ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_creatinine_umol_invalid(self):
        self.cleaned_data.update(
            creatinine=217,
            creatinine_units=MICROMOLES_PER_LITER,
            are_results_normal=YES,
        )
        form_validator = SpecimenResultFormValidator(
            cleaned_data=self.cleaned_data, instance=SpecimenResult()
        )
        self.assertRaises(ValidationError, form_validator.validate)
        self.assertIn("creatinine", form_validator._errors)

    def test_creatinine_umol(self):

        self.cleaned_data.update(creatinine=100, creatinine_units=MICROMOLES_PER_LITER)
        form_validator = SpecimenResultFormValidator(
            cleaned_data=self.cleaned_data, instance=SpecimenResult()
        )
        try:
            form_validator.validate()
        except ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got{e}")

    def test_magnesium_invalid(self):
        self.cleaned_data.update(magnesium=0.01, are_results_normal=YES)
        form_validator = SpecimenResultFormValidator(
            cleaned_data=self.cleaned_data, instance=SpecimenResult()
        )
        self.assertRaises(ValidationError, form_validator.validate)
        self.assertIn("magnesium", form_validator._errors)

    def test_magnesium(self):
        self.cleaned_data.update(magnesium=0.35, results_abnormal=YES, results_reportable=YES)
        form_validator = SpecimenResultFormValidator(
            cleaned_data=self.cleaned_data, instance=SpecimenResult()
        )
        self.assertRaises(ValidationError, form_validator.validate)
        self.assertIn("magnesium", form_validator._errors)

    def test_potassium_invalid(self):
        self.cleaned_data.update(potassium=1.0, results_abnormal=YES)
        form_validator = SpecimenResultFormValidator(
            cleaned_data=self.cleaned_data, instance=SpecimenResult()
        )
        self.assertRaises(ValidationError, form_validator.validate)
        self.assertIn("potassium", form_validator._errors)

    def test_potassium_high(self):

        self.cleaned_data.update(potassium=6.8, results_abnormal=YES, results_reportable=NO)
        form_validator = SpecimenResultFormValidator(
            cleaned_data=self.cleaned_data, instance=SpecimenResult()
        )
        self.assertRaises(ValidationError, form_validator.validate)
        self.assertIn("potassium", form_validator._errors)

    def test_potassium_low(self):
        self.cleaned_data.update(potassium=2.3, results_abnormal=YES, results_reportable=NO)
        form_validator = SpecimenResultFormValidator(
            cleaned_data=self.cleaned_data, instance=SpecimenResult()
        )
        self.assertRaises(ValidationError, form_validator.validate)
        self.assertIn("potassium", form_validator._errors)

    def test_sodium_invalid(self):
        self.cleaned_data.update(sodium=100, results_abnormal=YES)
        form_validator = SpecimenResultFormValidator(
            cleaned_data=self.cleaned_data, instance=SpecimenResult()
        )
        self.assertRaises(ValidationError, form_validator.validate)
        self.assertIn("sodium", form_validator._errors)

    def test_sodium_invalid_1(self):
        self.cleaned_data.update(
            sodium=119, results_abnormal=YES, results_reportable=NOT_APPLICABLE
        )
        form_validator = SpecimenResultFormValidator(
            cleaned_data=self.cleaned_data, instance=SpecimenResult()
        )
        self.assertRaises(ValidationError, form_validator.validate)
        self.assertIn("sodium", form_validator._errors)

    def test_sodium_invalid_2(self):
        self.cleaned_data.update(sodium=119, results_abnormal=YES, results_reportable=NO)
        form_validator = SpecimenResultFormValidator(
            cleaned_data=self.cleaned_data, instance=SpecimenResult()
        )
        self.assertRaises(ValidationError, form_validator.validate)
        self.assertIn("sodium", form_validator._errors)

    def test_sodium(self):
        self.cleaned_data.update(
            sodium=135, results_abnormal=NO, results_reportable=NOT_APPLICABLE
        )
        form_validator = SpecimenResultFormValidator(
            cleaned_data=self.cleaned_data, instance=SpecimenResult()
        )
        try:
            form_validator.validate()
        except ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got{e}")

    def test_alt_invalid(self):
        self.cleaned_data.update(alt=201, are_results_normal=YES)
        form_validator = SpecimenResultFormValidator(
            cleaned_data=self.cleaned_data, instance=SpecimenResult()
        )
        self.assertRaises(ValidationError, form_validator.validate)
        self.assertIn("alt", form_validator._errors)

    def test_alt(self):

        self.cleaned_data.update(alt=10, results_abnormal=NO)
        form_validator = SpecimenResultFormValidator(
            cleaned_data=self.cleaned_data, instance=SpecimenResult()
        )
        try:
            form_validator.validate()
        except ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got{e}")

    def test_platelets_invalid(self):

        self.cleaned_data.update(platelets=50, results_abnormal=YES)
        form_validator = SpecimenResultFormValidator(
            cleaned_data=self.cleaned_data, instance=SpecimenResult()
        )
        self.assertRaises(ValidationError, form_validator.validate)
        self.assertIn("platelets", form_validator._errors)

    def test_platelets(self):

        self.cleaned_data.update(platelets=450, results_abnormal=NO)
        form_validator = SpecimenResultFormValidator(
            cleaned_data=self.cleaned_data, instance=SpecimenResult()
        )
        try:
            form_validator.validate()
        except ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_neutrophil_invalid(self):
        self.cleaned_data.update(neutrophil=0.5)
        form_validator = SpecimenResultFormValidator(
            cleaned_data=self.cleaned_data, instance=SpecimenResult()
        )
        self.assertRaises(ValidationError, form_validator.validate)
        self.assertIn("neutrophil", form_validator._errors)

    def test_neutrophil(self):

        self.cleaned_data.update(neutrophil=4, results_abnormal=NO)
        form_validator = SpecimenResultFormValidator(
            cleaned_data=self.cleaned_data, instance=SpecimenResult()
        )
        try:
            form_validator.validate()
        except ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got{e}")

    def test_results_reportable_invalid(self):
        self.cleaned_data.update(sodium=1000, results_abnormal=YES, results_reportable=NO)
        form_validator = SpecimenResultFormValidator(
            cleaned_data=self.cleaned_data, instance=SpecimenResult()
        )
        self.assertRaises(ValidationError, form_validator.validate)
        self.assertIn("sodium", form_validator._errors)



from django import forms
from edc_constants.constants import NO, NOT_APPLICABLE, YES
from edc_metadata.constants import REQUIRED

from .constants import ALREADY_REPORTED, INVALID_REFERENCE, PRESENT_AT_BASELINE
from .site_reportables import site_reportables
from .value_reference_group import NotEvaluated


class UserResponse:
    def __init__(self, utest_id, cleaned_data=None):
        # ensure each supporting option is provided from the form
        for attr in ["units", "abnormal", "reportable"]:
            if not cleaned_data.get(f"{utest_id}_{attr}"):
                raise forms.ValidationError(
                    {f"{utest_id}_{attr}": "This field is required."}, code=REQUIRED
                )

        self.abnormal = cleaned_data.get(f"{utest_id}_abnormal")
        self.reportable = cleaned_data.get(f"{utest_id}_reportable")
        self.units = cleaned_data.get(f"{utest_id}_units")


class ReportablesEvaluator:
    def __init__(
        self,
        reference_range_collection_name=None,
        cleaned_data=None,
        gender=None,
        dob=None,
        report_datetime=None,
        value_field_suffix=None,
        **extra_options,
    ):
        self.reference_range_collection = site_reportables.get(reference_range_collection_name)
        if not self.reference_range_collection:
            raise forms.ValidationError(
                {
                    "Invalid reference range collection. "
                    f"Got '{reference_range_collection_name}'"
                },
                code=INVALID_REFERENCE,
            )

        self.cleaned_data = cleaned_data
        self.dob = dob
        self.gender = gender
        self.report_datetime = report_datetime
        self.value_field_suffix = value_field_suffix
        self.extra_options = extra_options

    def validate_reportable_fields(self):
        """Check normal ranges and grade result values
        for each field mentioned in the reference_range_collection.
        """
        for field, value in self.cleaned_data.items():
            try:
                utest_id, _ = field.split(self.value_field_suffix)
            except ValueError:
                utest_id = field
            if value is not None and self.reference_range_collection.get(utest_id):
                self._evaluate_reportable(utest_id, value, field)

    def validate_results_abnormal_field(
        self, field=None, responses=None, suffix=None, word=None
    ):
        """Validate the "results_abnormal" field."""
        self._validate_final_assessment(
            field=field or "results_abnormal",
            responses=responses or [YES],
            suffix=suffix or "_abnormal",
            word=word or "abnormal",
        )

    def validate_results_reportable_field(
        self, field=None, responses=None, suffix=None, word=None
    ):
        """Validate the "results_reportable" field."""
        self._validate_final_assessment(
            field=field or "results_reportable",
            responses=responses,
            suffix=suffix or "_reportable",
            word=word or "reportable",
        )

    def reportable_grades_for_utest_id(self, utest_id):
        try:
            reportable_grades = (
                self.reference_range_collection.reportable_grades_exceptions.get(utest_id)
            )
        except KeyError:
            pass
        return reportable_grades or self.reference_range_collection.reportable_grades

    def _evaluate_reportable(self, utest_id, value, field):
        """Evaluate a single result value.

        Grading is done first. If the value is not gradeable,
        the value is checked against the normal limits.

        Expected field naming convention:
            * {utest_id}_value
            * {utest_id}_units
            * {utest_id}_abnormal [YES, (NO)]
            * {utest_id}_reportable [(NOT_APPLICABLE), NO, GRADE3, GRADE4]
        """

        # get relevant user form reponses
        response = UserResponse(utest_id, self.cleaned_data)

        opts = dict(
            dob=self.dob,
            gender=self.gender,
            report_datetime=self.report_datetime,
            units=response.units,
            **self.extra_options,
        )

        grade = self.get_grade(utest_id, value, field, **opts)

        # is gradeable, user reponse matches grade or has valid opt out
        # response
        if (
            grade
            and grade.grade
            and str(grade.grade) in self.reportable_grades_for_utest_id(utest_id)
            and response.reportable
            not in [str(grade.grade), ALREADY_REPORTED, PRESENT_AT_BASELINE]
        ):
            raise forms.ValidationError(
                {field: f"{utest_id.upper()} is reportable. Got {grade.description}."}
            )

        # is not gradeable, user reponse is a valid opt out
        if not grade and response.reportable not in [NO, NOT_APPLICABLE]:
            raise forms.ValidationError(
                {f"{utest_id}_reportable": "Invalid. Expected 'No' or 'Not applicable'."}
            )

        normal = self.get_normal(utest_id, value, field, **opts)

        # is not normal, user response does not match
        if not normal and response.abnormal == NO:
            descriptions = self.reference_range_collection.get(
                utest_id
            ).get_normal_description(**opts)
            raise forms.ValidationError(
                {
                    field: (
                        f"{utest_id.upper()} is abnormal. "
                        f"Normal ranges: {', '.join(descriptions)}"
                    )
                }
            )

        # is not normal and not gradeable, user response does not match
        if normal and not grade and response.abnormal == YES:
            raise forms.ValidationError(
                {f"{utest_id}_abnormal": "Invalid. Result is not abnormal"}
            )

        # illogical user response combination
        if response.abnormal == YES and response.reportable == NOT_APPLICABLE:
            raise forms.ValidationError(
                {f"{utest_id}_reportable": "This field is applicable if result is abnormal"}
            )

        # illogical user response combination
        if response.abnormal == NO and response.reportable != NOT_APPLICABLE:
            raise forms.ValidationError(
                {f"{utest_id}_reportable": "This field is not applicable"}
            )

    def get_grade(self, utest_id=None, value=None, field=None, **opts):
        try:
            grade = self.reference_range_collection.get(utest_id).get_grade(value, **opts)
        except NotEvaluated as e:
            raise forms.ValidationError({field: str(e)})
        return grade

    def get_normal(self, utest_id=None, value=None, field=None, **opts):
        try:
            normal = self.reference_range_collection.get(utest_id).get_normal(value, **opts)
        except NotEvaluated as e:
            raise forms.ValidationError({field: str(e)})
        return normal

    def _validate_final_assessment(self, field=None, responses=None, suffix=None, word=None):
        """Common code to validate fields `results_abnormal`
        and `results_reportable`.
        """
        responses = responses or self.reference_range_collection.reportable_grades
        answers = list(
            {k: v for k, v in self.cleaned_data.items() if k.endswith(suffix)}.values()
        )
        if len([True for v in answers if v is not None]) == 0:
            raise forms.ValidationError({"results_abnormal": "No results have been entered."})
        answers_as_bool = [True for v in answers if v in responses]
        if self.cleaned_data.get(field) == NO:
            if any(answers_as_bool):
                are = "is" if len(answers_as_bool) == 1 else "are"
                raise forms.ValidationError(
                    {field: f"{len(answers_as_bool)} of the above results {are} {word}"}
                )
        elif self.cleaned_data.get(field) == YES:
            if not any(answers_as_bool):
                raise forms.ValidationError({field: f"None of the above results are {word}"})



from django.test import TestCase

from edc_reportable import convert_units
from edc_reportable.units import (
    MICROMOLES_PER_LITER,
    MILLIGRAMS_PER_DECILITER,
    MILLIMOLES_PER_LITER,
)


class TestParser(TestCase):
    def test_convert_glucose(self):
        """mg/dL to mmol/L"""

        values = [
            (1.0, 18.0180),
            (19.0, 342.3423),
            (33.0, 594.5946),
            (37.0, 666.6667),
            (125.0, 6.9375),
        ]

        for value, converted_value in values:
            converted_value = convert_units(
                value,
                units_from=MILLIGRAMS_PER_DECILITER,
                units_to=MILLIMOLES_PER_LITER,
            )
            self.assertEqual(converted_value, converted_value)

        converted_value = convert_units(
            558.558559, units_from=MILLIMOLES_PER_LITER, units_to=MILLIMOLES_PER_LITER
        )
        self.assertEqual(558.5586, converted_value)

        converted_value = convert_units(
            6.9375, units_from=MILLIMOLES_PER_LITER, units_to=MILLIMOLES_PER_LITER
        )
        self.assertEqual(6.9375, converted_value)

    def test_convert_creatinine(self):
        """mg/dL to umol/L"""

        values = [
            (1.0, 0.0113),
            (19.0, 0.2149),
            (33.0, 0.3733),
            (37.0, 0.4186),
            (50.0, 0.5656),
        ]

        for value, converted_value in values:
            converted_value = convert_units(
                value,
                units_from=MILLIGRAMS_PER_DECILITER,
                units_to=MICROMOLES_PER_LITER,
            )
            self.assertEqual(converted_value, converted_value)

        converted_value = convert_units(
            0.2149, units_from=MICROMOLES_PER_LITER, units_to=MICROMOLES_PER_LITER
        )
        self.assertEqual(0.2149, converted_value)



from copy import copy
from datetime import datetime

from dateutil.relativedelta import relativedelta
from django.test import TestCase
from edc_constants.constants import FEMALE, MALE
from edc_utils import get_utcnow
from pytz import utc

from edc_reportable import (
    IU_LITER,
    BoundariesOverlap,
    GradeError,
    GradeReference,
    NormalReference,
    NotEvaluated,
    ValueReferenceGroup,
    adult_age_options,
)
from edc_reportable.constants import HIGH_VALUE


class TestGrading(TestCase):
    def test_grading(self):
        dob = get_utcnow() - relativedelta(years=25)
        report_datetime = utc.localize(datetime(2017, 12, 7))
        grp = ValueReferenceGroup(name="labtest")

        opts = dict(
            name="labtest",
            grade=2,
            lower=10,
            upper=20,
            units="mg/dL",
            age_lower=18,
            age_upper=99,
            age_units="years",
            gender=MALE,
        )

        g2 = GradeReference(**opts)
        self.assertTrue(repr(g2))

        new_opts = copy(opts)
        new_opts.update(grade="-1")
        self.assertRaises(GradeError, GradeReference, **new_opts)

        for grade in range(0, 6):
            new_opts = copy(opts)
            new_opts.update(grade=str(grade))
            try:
                GradeReference(**new_opts)
            except GradeError:
                self.fail("GradeError unexpectedly raised")

        new_opts = copy(opts)
        new_opts.update(grade=3, lower=20, lower_inclusive=True, upper=30)
        g3 = GradeReference(**new_opts)

        new_opts = copy(opts)
        new_opts.update(grade=4, lower=30, lower_inclusive=True, upper=40)
        g4 = GradeReference(**new_opts)

        grp.add_grading(g2)
        grp.add_grading(g3)
        grp.add_grading(g4)

        grade = grp.get_grade(
            value=10,
            gender=MALE,
            dob=dob,
            report_datetime=report_datetime,
            units="mg/dL",
        )
        self.assertIsNone(grade)
        grade = grp.get_grade(
            value=11,
            gender=MALE,
            dob=dob,
            report_datetime=report_datetime,
            units="mg/dL",
        )
        self.assertEqual(grade.grade, 2)

        grade = grp.get_grade(
            value=20,
            gender=MALE,
            dob=dob,
            report_datetime=report_datetime,
            units="mg/dL",
        )
        self.assertEqual(grade.grade, 3)
        grade = grp.get_grade(
            value=21,
            gender=MALE,
            dob=dob,
            report_datetime=report_datetime,
            units="mg/dL",
        )
        self.assertEqual(grade.grade, 3)

        grade = grp.get_grade(
            value=30,
            gender=MALE,
            dob=dob,
            report_datetime=report_datetime,
            units="mg/dL",
        )
        self.assertEqual(grade.grade, 4)
        grade = grp.get_grade(
            value=31,
            gender=MALE,
            dob=dob,
            report_datetime=report_datetime,
            units="mg/dL",
        )
        self.assertEqual(grade.grade, 4)

        self.assertRaises(
            NotEvaluated,
            grp.get_grade,
            value=31,
            gender=MALE,
            dob=report_datetime.date(),
            report_datetime=report_datetime,
            units="mg/dL",
        )

        self.assertRaises(
            NotEvaluated,
            grp.get_grade,
            value=31,
            gender=MALE,
            dob=dob,
            report_datetime=report_datetime,
            units="mmol/L",
        )

        self.assertRaises(
            NotEvaluated,
            grp.get_grade,
            value=31,
            gender=FEMALE,
            dob=dob,
            report_datetime=report_datetime,
            units="mmol/L",
        )

        grade = grp.get_grade(
            value=1,
            gender=MALE,
            dob=dob,
            report_datetime=report_datetime,
            units="mg/dL",
        )
        self.assertIsNone(grade)

        new_opts = copy(opts)
        new_opts.update(grade=1, lower=15, upper=20)

        # overlaps with G2
        g1 = GradeReference(**new_opts)
        grp.add_grading(g1)

        self.assertRaises(
            BoundariesOverlap,
            grp.get_grade,
            value=16,
            gender=MALE,
            dob=dob,
            report_datetime=report_datetime,
            units="mg/dL",
        )

    def test_grading_with_limits_normal(self):
        dob = get_utcnow() - relativedelta(years=25)
        report_datetime = utc.localize(datetime(2017, 12, 7))
        grp = ValueReferenceGroup(name="amylase")
        normal_reference = NormalReference(
            name="amylase",
            gender=[MALE, FEMALE],
            units=IU_LITER,
            lower=25.0,
            upper=125.0,
            lower_inclusive=True,
            upper_inclusive=True,
            **adult_age_options,
        )
        opts = dict(
            name="amylase",
            grade=1,
            lower="1.1*ULN",
            upper="1.5*ULN",
            lower_inclusive=True,
            upper_inclusive=False,
            units=IU_LITER,
            gender=MALE,
            **adult_age_options,
        )
        g1 = GradeReference(normal_references={MALE: [normal_reference]}, **opts)

        new_opts = copy(opts)
        new_opts.update(grade=2, lower="1.5*ULN", upper="3.0*ULN")
        g2 = GradeReference(normal_references={MALE: [normal_reference]}, **new_opts)

        new_opts = copy(opts)
        new_opts.update(grade=3, lower="3.0*ULN", upper="5.0*ULN")
        g3 = GradeReference(normal_references={MALE: [normal_reference]}, **new_opts)

        new_opts = copy(opts)
        new_opts.update(grade=4, lower="5.0*ULN", upper=f"{HIGH_VALUE}*ULN")
        g4 = GradeReference(normal_references={MALE: [normal_reference]}, **new_opts)

        grp.add_grading(g1)
        grp.add_grading(g2)
        grp.add_grading(g3)
        grp.add_grading(g4)

        grade = grp.get_grade(
            value=130,
            gender=MALE,
            dob=dob,
            report_datetime=report_datetime,
            units=IU_LITER,
        )
        self.assertIsNone(grade)

        grade = grp.get_grade(
            value=137.5,
            gender=MALE,
            dob=dob,
            report_datetime=report_datetime,
            units=IU_LITER,
        )
        self.assertEqual(grade.grade, 1)

        grade = grp.get_grade(
            value=187.4,
            gender=MALE,
            dob=dob,
            report_datetime=report_datetime,
            units=IU_LITER,
        )
        self.assertEqual(grade.grade, 1)

        grade = grp.get_grade(
            value=187.5,
            gender=MALE,
            dob=dob,
            report_datetime=report_datetime,
            units=IU_LITER,
        )
        self.assertEqual(grade.grade, 2)

        grade = grp.get_grade(
            value=212,
            gender=MALE,
            dob=dob,
            report_datetime=report_datetime,
            units=IU_LITER,
        )
        self.assertEqual(grade.grade, 2)

        grade = grp.get_grade(
            value=600,
            gender=MALE,
            dob=dob,
            report_datetime=report_datetime,
            units=IU_LITER,
        )
        self.assertEqual(grade.grade, 3)

        grade = grp.get_grade(
            value=780,
            gender=MALE,
            dob=dob,
            report_datetime=report_datetime,
            units=IU_LITER,
        )
        self.assertEqual(grade.grade, 4)



from .units import MICROMOLES_PER_LITER, MILLIGRAMS_PER_DECILITER, MILLIMOLES_PER_LITER


class ConversionNotHandled(Exception):
    pass


def convert_units(value, units_from=None, units_to=None, places=None):
    places = places or 4
    converted_value = value
    if value and units_from and units_to and units_from != units_to:
        if (units_from, units_to) == (MILLIGRAMS_PER_DECILITER, MILLIMOLES_PER_LITER):
            converted_value = float(value) / 18.018018
        elif (units_from, units_to) == (MILLIMOLES_PER_LITER, MILLIGRAMS_PER_DECILITER):
            converted_value = float(value) * 18.018018
        elif (units_from, units_to) == (MILLIGRAMS_PER_DECILITER, MICROMOLES_PER_LITER):
            converted_value = float(value) * 88.42
        elif (units_from, units_to) == (MICROMOLES_PER_LITER, MILLIGRAMS_PER_DECILITER):
            converted_value = float(value) / 88.42
        else:
            raise ConversionNotHandled(
                f"Conversion not handled. Got from {units_from} to {units_to}"
            )
    if converted_value:
        converted_value = round(converted_value, places)
    return converted_value



from edc_utils import get_utcnow

GRADING = "grading"
NORMAL = "normal"


def get_references(
    value_references=None,
    gender=None,
    dob=None,
    report_datetime=None,
    units=None,
    **extra_options,
):
    """Returns a list of references for this
    gender, age and units.

    Either ValueReferences or GradeReferences.
    """
    references = []
    report_datetime = report_datetime or get_utcnow()
    for refs in value_references.values():
        for ref in refs:
            non_matching_opts = [k for k, v in extra_options.items() if getattr(ref, k) != v]
            if (
                gender in ref.gender
                and ref.units == units
                and ref.age_match(dob, report_datetime)
                and not non_matching_opts
            ):
                references.append(ref)
    return references


class InvalidValueReference(Exception):
    pass


class ValueReferenceNotFound(Exception):
    pass


class ValueReferenceAlreadyAdded(Exception):
    pass


class BoundariesOverlap(Exception):
    pass


class NotEvaluated(Exception):
    pass


class Result:
    def __init__(self, value, description):
        self._value = value
        self.description = description(value=value)

    def __str__(self):
        return self.description


class Normal(Result):
    pass


class Grade(Result):
    def __init__(self, value, grade, description):
        super().__init__(value, description)
        self.grade = grade


class ValueReferenceGroup:
    def __init__(self, name=None):
        self.name = name
        self.normal = {}
        self.grading = {}

    def __repr__(self):
        return f"{self.__class__.__name__}(name={self.name})"

    def add_normal(self, normal_reference):
        """Adds a ValueReference to the dictionary of
        normal references.
        """
        self._add(normal_reference, self.normal)

    def add_grading(self, grade_reference):
        """Adds a GradeReference to the dictionary of
        grading references.
        """
        self._add(grade_reference, self.grading)

    def get_normal_description(self, **kwargs):
        """Returns a list of descriptions of the normal references
        for these criteria.
        """
        descriptions = []
        for value_ref in self._get_normal_references(**kwargs):
            descriptions.append(value_ref.description())
        return descriptions

    def get_normal(self, value=None, **kwargs):
        """Returns a Normal instance or None."""
        normal = None
        for value_ref in self._get_normal_references(**kwargs):
            if value_ref.in_bounds(value=value, **kwargs):
                if not normal:
                    normal = Normal(value, value_ref.description)
                else:
                    raise BoundariesOverlap(
                        f"Previously got {normal}. "
                        f"Got {value_ref.description(value=value)}. "
                        "Check your definitions."
                    )
        return normal

    def get_grade(self, value=None, **kwargs):
        """Returns a Grade instance or None."""
        grade = None
        for grade_ref in self._get_grading_references(**kwargs):
            if grade_ref.in_bounds(value=value, **kwargs):
                if not grade:
                    grade = Grade(value, grade_ref.grade, grade_ref.description)
                else:
                    raise BoundariesOverlap(
                        f"Previously got {grade}. "
                        f"Got {grade_ref.description(value=value)} ",
                        "Check your definitions.",
                    )
        return grade

    def _get_normal_references(self, **kwargs):
        """Returns a list of ValueReference instances or raises."""
        references = get_references(value_references=self.normal, **kwargs)
        if not references:
            raise NotEvaluated(
                f"{self.name} value not evaluated. "
                f"No reference range found for {kwargs}. See {repr(self)}."
            )
        return references

    def _get_grading_references(self, **kwargs):
        """Returns a list of GradeReference instances or raises."""
        references = get_references(value_references=self.grading, **kwargs)
        if not references:
            raise NotEvaluated(
                f"{self.name} value not graded. "
                f"No reference range found for {kwargs}. See {repr(self)}."
            )
        references.sort(key=lambda x: x.grade, reverse=True)
        return references

    def _add(self, value_reference, value_references):
        if value_reference.name != self.name:
            raise InvalidValueReference(
                "Cannot add to group; name does not match. "
                f"Expected '{self.name}'. Got '{value_reference.name}'. "
                f"See {repr(value_reference)}"
            )
        try:
            if value_reference in value_references[value_reference.gender]:
                raise ValueReferenceAlreadyAdded(
                    f"Value reference {value_reference} has already been added."
                )
        except KeyError:
            value_references[value_reference.gender] = []
        value_references[value_reference.gender].append(value_reference)



from datetime import datetime

from dateutil.relativedelta import relativedelta
from django.test import TestCase
from edc_constants.constants import FEMALE, MALE
from edc_utils import get_utcnow
from pytz import utc

from edc_reportable import (
    AlreadyRegistered,
    GradeReference,
    NormalReference,
    ReferenceRangeCollection,
    ValueReferenceGroup,
)


class TestCollection(TestCase):
    def test_collection(self):
        dob = get_utcnow() - relativedelta(years=25)
        report_datetime = utc.localize(datetime(2017, 12, 7))
        reference = ReferenceRangeCollection()

        self.assertTrue(repr(reference))

        neutrophils = ValueReferenceGroup(name="neutrophils")
        ln = NormalReference(
            name="neutrophils",
            lower=2.5,
            upper=7.5,
            units="10e9/L",
            age_lower=18,
            age_upper=99,
            age_units="years",
            gender=[MALE, FEMALE],
        )

        g3 = GradeReference(
            name="neutrophils",
            grade=3,
            lower=0.4,
            lower_inclusive=True,
            upper=0.59,
            upper_inclusive=True,
            units="10e9/L",
            age_lower=18,
            age_upper=99,
            age_units="years",
            gender=[MALE, FEMALE],
        )

        g4 = GradeReference(
            name="neutrophils",
            grade=4,
            lower=0,
            upper=0.4,
            units="10e9/L",
            age_lower=18,
            age_upper=99,
            age_units="years",
            gender=[MALE, FEMALE],
        )

        neutrophils.add_normal(ln)
        neutrophils.add_grading(g3)
        neutrophils.add_grading(g4)

        reference.register(neutrophils)
        self.assertRaises(AlreadyRegistered, reference.register, neutrophils)
        self.assertTrue(
            reference.get("neutrophils").get_normal(
                value=3.5,
                units="10e9/L",
                gender=MALE,
                dob=dob,
                report_datetime=report_datetime,
            )
        )

        neutrophils = reference.get("neutrophils")
        self.assertTrue(
            neutrophils.get_normal(
                value=3.5,
                units="10e9/L",
                gender=MALE,
                dob=dob,
                report_datetime=report_datetime,
            )
        )
        grade = neutrophils.get_grade(
            value=3.5,
            units="10e9/L",
            gender=MALE,
            dob=dob,
            report_datetime=report_datetime,
        )
        self.assertIsNone(grade)

        grade = neutrophils.get_grade(
            value=0.43,
            units="10e9/L",
            gender=MALE,
            dob=dob,
            report_datetime=report_datetime,
        )
        self.assertEqual(grade.grade, 3)

        grade = neutrophils.get_grade(
            value=0.3,
            units="10e9/L",
            gender=MALE,
            dob=dob,
            report_datetime=report_datetime,
        )
        self.assertEqual(grade.grade, 4)



from .constants import LLN, ULN
from .value_reference import ValueReference


class NormalReferenceError(Exception):
    pass


class NormalReference(ValueReference):
    def __init__(self, name=None, lower=None, upper=None, **kwargs):
        string = f"{(lower or '')}{(upper or '')}"
        if ULN in string or LLN in string:
            raise NormalReferenceError(
                f"{LLN} and {ULN} are not relevant to a normal range. See {name}."
            )
        super().__init__(name=name, lower=lower, upper=upper, **kwargs)



import re
from collections import OrderedDict

from edc_constants.constants import FEMALE, MALE

from .adult_age_options import adult_age_options
from .constants import LLN, ULN


class ParserError(Exception):
    pass


def unparse(**kwargs):

    lower = kwargs.get("lower") or ""
    upper = kwargs.get("upper") or ""
    lower_op = "" if not lower else "<=" if kwargs.get("lower_inclusive") else "<"
    upper_op = "" if not upper else "<=" if kwargs.get("upper_inclusive") else "<"
    gender = kwargs.get("gender", "")
    age_lower = kwargs.get("age_lower", "")
    age_upper = kwargs.get("age_upper", "")
    age_lower_op = "" if not age_lower else "<=" if kwargs.get("age_lower_inclusive") else "<"
    age_upper_op = "" if not age_upper else "<=" if kwargs.get("age_upper_inclusive") else "<"
    age = (
        ""
        if not age_lower and not age_upper
        else f"{age_lower}{age_lower_op}AGE{age_upper_op}{age_upper}"
    )
    try:
        fasting = kwargs.pop("fasting")
    except KeyError:
        fasting_str = ""
    else:
        fasting_str = "Fasting " if fasting else ""
    return f"{lower}{lower_op}x{upper_op}{upper} {fasting_str}{gender} {age}".rstrip()


def parse(phrase=None, fasting=None, uln=None, lln=None, **kwargs):

    pattern = r"(([\d+\.\d+]|[\.\d+])?(<|<=)?)+x((<|<=)?([\d+\.\d+]|[\.\d+])+)?"
    lln = f"*{LLN}"
    uln = f"*{ULN}"

    def _parse(string):
        inclusive = True if "=" in string else None
        try:
            value = float(
                string.replace("<", "").replace("=", "").replace(lln, "").replace(uln, "")
            )
        except ValueError:
            value = None
        if lln in string:
            value = f"{value}{lln}"
        elif uln in string:
            value = f"{value}{uln}"
        return value, inclusive

    phrase = phrase.replace(" ", "")
    match = re.match(pattern, phrase.replace(lln, "").replace(uln, ""))
    if not match or match.group() != phrase.replace(lln, "").replace(uln, ""):
        raise ParserError(
            f"Invalid. Got {phrase}. Expected, e.g, 11<x<22, "
            "11<=x<22, 11<x<=22, 11<x, 11<=x, x<22, x<=22, etc."
        )
    left, right = phrase.replace(" ", "").split("x")
    lower, lower_inclusive = _parse(left)
    upper, upper_inclusive = _parse(right)
    fasting = True if fasting else False
    ret = OrderedDict(
        lower=lower,
        lower_inclusive=lower_inclusive,
        upper=upper,
        upper_inclusive=upper_inclusive,
        fasting=fasting,
        **kwargs,
    )
    for k, v in ret.items():
        setattr(ret, k, v)
    return ret


def parse_boundary(value):
    uln = lln = None
    value = value.upper()
    pattern = r"(\d+\.\d+)\*[LLN|ULN]"
    if not re.match(pattern, value):
        raise ParserError(f"Invalid value. Got {value}")
    else:
        value, limit_normal = value.split("*")
        if limit_normal == LLN:
            lln = float(value)
        elif limit_normal == ULN:
            uln = float(value)
    return lln, uln


def dummy_parse(phrase=None, units=None):
    return parse(
        phrase or "0<=x",
        units=units,
        gender=[MALE, FEMALE],
        **adult_age_options,
    )
